#include "example.h"
#include <string>

Alphabet Example :: read_alphabet() {
  string input;
  Alphabet result;

  cout << "Please enter each symbol in the alphabet, together with its "
    "type, in the format f : (o -> o) -> o.  Finish with an empty line."
    << endl << "> ";
  
  getline(cin, input);
  while (input != "") {
    // find separating colon
    int colon = input.find(':');
    if (colon == string::npos) {
      cout << "Illegal format; missing colon." << endl << "> ";
      getline(cin, input);
      continue;
    }
    // is the thing before the colon a single word?
    string name = input.substr(0,colon);
    while (name.length() > 0 && name[0] == ' ') name = name.substr(1);
    while (name.length() > 0 && name[name.length()-1] == ' ')
      name = name.substr(0,name.length()-1);
    if (name.length() == 0) {
      cout << "Illegal format: missing constant name." << endl << "> ";
      getline(cin, input);
      continue;
    }
    if (name.find(' ') != string::npos) {
      cout << "Illegal format: constant name should be a single word." << endl;
      cout << "> ";
      getline(cin, input);
      continue;
    }
    // is the thing after it a legal type?
    string typetxt = input.substr(colon+1);
    PType type = TYPE(typetxt);
    if (type == 0) {
      getline(cin, input);
      continue;
    }

    // all good! Add it to the alphabet :)
    result.add(new Constant(name, type));

    cout << "> ";
    getline(cin, input);
  }
  return result;
}

Ruleset Example :: read_rules(Alphabet sigma) {
  string input;
  Ruleset result;

  cout << "Please enter every rule, in the format f (/\\x:o. g x Y) => h Y.  "
    "Finish with an empty line." << endl << "> ";

  getline(cin, input);
  while (input != "") {
    // find separating arrow
    int arrow = input.find("=>");
    if (arrow == string::npos) {
      cout << "Illegal format, missing =>." << endl << "> ";
      getline(cin, input);
      continue;
    }

    // can we make terms out of both sides?
    string left = input.substr(0, arrow);
    string right = input.substr(arrow+2);
    Environment gamma;
    PTerm l = TERM(left, sigma, gamma);
    if (l == NULL) {
      cout << "Error in left hand side " << left << "." << endl << "> ";
      getline(cin, input);
      continue;
    }
    PTerm r = TERM(right, sigma, gamma);
    if (r == NULL) {
      delete l;
      cout << "Error in right hand side " << right << "." << endl << "> ";
      getline(cin, input);
      continue;
    }

    // can we make it a rule?
    MatchRule* rule = new MatchRule(gamma, l, r);
    if (!rule->query_valid()) {
      cout << "Invalid rule: " << rule->query_invalid_reason() << "." << endl
        << "> ";
      rule->free_memory();
      getline(cin, input);
      continue;
    }
    // we can!
    result.push_back(rule);

    cout << "> ";
    getline(cin, input);
  }
  result.push_back(new Beta);
  return result;
}

PTerm Example :: read_term(Alphabet sigma) {
  string term;
  Environment gamma;

  cout << "Please enter a term you would like to see normalised.  The type "
    "may be included (in the form term : type) or omitted if it is evident."
    << endl << "> ";

  getline(cin, term);
  int colon = term.find_last_of(':');
  if (colon == string::npos || term.find('.', colon+1) != string::npos) {
    return TERM(term, sigma, gamma);
    // gamma will just be FV(term), so can be forgotten about
  }
  else {
    PType type = TYPE(term.substr(colon+1));
    if (type == NULL) return NULL;
    PTerm t = TERM(term.substr(0,colon), sigma, gamma, type);
    delete type;
    return t;
  }
}

bool Example :: applicable_rule(PTerm term, Ruleset rules) {
  if (term == NULL) return false;

  Ruleset::iterator it;
  for (it = rules.begin(); it != rules.end(); it++) {
    if ((*it)->applicable(term, "")) return true;
  }
  return false;
}

list<string> Example :: find_reducable_positions(PTerm term, Ruleset rules) {
  list<string> positions = term->query_positions();
  list<string> ret;
  while (!positions.empty()) {
    if (applicable_rule(term->subterm(positions.front()), rules)) {
      ret.push_back(positions.front());
    }
    positions.pop_front();
  }
  return ret;
}

string Example :: random_position(list<string> positions) {
  int N = random() % positions.size();
  for (int i = 0; i < N; i++) {
    positions.push_back(positions.front());
    positions.pop_front();
  }
  return positions.front();
}

PTerm Example :: reduce_on_position(PTerm term, string position,
                                    Ruleset rules) {
  int N = random() % rules.size();
  int i;
  for (i = 0; i <= N; ) {
    rules.push_back(rules.front());
    rules.pop_front();
    if (rules.front()->applicable(term, position)) {
      i++;
    }
  }
  return rules.front()->apply(term, position);
}

int Example :: main() {
  Alphabet sigma = read_alphabet();
  Ruleset rules = read_rules(sigma);
  PTerm termpje = NULL;
  while (termpje == NULL) termpje = read_term(sigma);

  cout << termpje->to_string() << endl;
  list<string> positions = find_reducable_positions(termpje, rules);
  while (positions.size() > 0) {
    termpje = reduce_on_position(termpje,
                                 random_position(positions),
                                 rules);
    cout << "  ==> " << termpje->to_string() << endl;
    positions = find_reducable_positions(termpje, rules);
  }
  return 0;
}


Example prog;

int main() {
  return prog.main();
}

