#ifndef ENVIRONMENT_H
#define ENVIRONMENT_H

/**
 * This class implements an environment of variables.
 *
 * Variables in a term are considered identical only if they are the
 * same pointer.  Hence, variables are never deleted, unless all
 * references to the variable have been removed.
 * For bound variables, the variable is remembered in the abstraction
 * that binds it.  Unbound variables are remembered in an
 * environment.
 *
 * As in the theory behind it, environments change depending on the
 * place in a term.  Roughly, if gamma |- (/\ x:o.C[x]) : o->a, then
 * gamma + {x:o} |- C[x]:a.  As such, environments can have elements
 * added and removed.
 *
 * A copy of an environment will contain pointers to the same memory
 * (a copy is not a deep copy), so only call free_memory() on an
 * Environment if its elements are no longer used anywhere, and no
 * shallow copies of the environment exist.
 */

#include <map>
#include "term.h"
#include "alphabet.h"

class Environment : public map<string,Variable*> {
  protected:
    string choose_new_name(Alphabet sigma, PType type, bool bound);

  public:
    void add(Variable* v);
    string add(PType type, Alphabet sigma, bool bound = true);
      // this will return a selfchosen name, by which the new
      // variable is added
    void remove(string v);
      // warning: caller needs to make sure the Variable* is deleted!
    bool contains(string key);
    void combine(Environment other);
      // adds all variables from other into this environment,
      // avoiding duplicates
      // variables are considered duplicates when they have the same
      // name; variables will not be renamed!
    void setminus(Environment other);
      // removes all variables from other out of this environment
      // we assume variables with the same name are the same;
      // pointers to variables are not deleted!
    bool subset(Environment other);
      // returns whether we are completely contained in other
      // (both keywise and actual variable wise)
    void free_memory();
    string to_string();
};

#endif

