#include "matchrule.h"
#include "substitution.h"

MatchRule :: MatchRule() : valid(false), l(NULL), r(NULL),
  invalid_reason("rule not set") {}

MatchRule :: MatchRule(Environment _gamma, PTerm _l, PTerm _r)
    : gamma(_gamma), l(_l), r(_r), valid(true) {
  check_valid();
}

void MatchRule :: set_valid(bool value) {
  valid = value;
}

bool MatchRule :: query_valid() {
  return valid;
}

void MatchRule :: check_valid() {
  invalid_reason = "rule is valid";

  // we must have a left hand side and a right hand side
  if (l == NULL || r == NULL) {
    if (l == NULL) invalid_reason = "left hand side not set up";
    else invalid_reason = "right hand side not set up";
    valid = false;
    return;
  }

  // the left hand side must be headed by a constant
  if (!l->query_head()->query_constant()) {
    invalid_reason = "left hand side should be headed by a constant";
    valid = false;
    return;
  }

  // both sides must have the same type
  if (!l->query_type()->equals(r->query_type())) {
    invalid_reason = "both sides should have the same type";
    valid = false;
    return;
  }

  // all variables from the left hand side need to occur in the environment
  Environment FVl = l->free_var();
  if (!FVl.subset(gamma)) {
    invalid_reason = "free variables in left hand side should occur in the "
      "given environment";
    valid = false;
    return;
  }

  // all variables from the right hand side need to occur in the left hand side
  Environment FVr = r->free_var();
  if (!FVr.subset(FVl)) {
    invalid_reason = "free variables in the right hand side should occur in "
      "the left hand side";
    valid = false;
    return;
  }
}

string MatchRule :: query_invalid_reason() {
  return invalid_reason;
}

bool MatchRule :: match_left(PTerm left, PTerm with, Substitution &match,
                             Environment bound) {

  /* variables */
  if (left->query_variable()) {
    PVariable ll = dynamic_cast<PVariable>(left);

    // bound variable, we need the same one in with
    if (bound.contains(ll->query_name())) return left == with;

    // free variable that is used elsewhere, should map to the same term
    if (match.contains(ll)) return match[ll]->equals(with);

    // unused variable; with may not contain bound variables
    Environment FV = with->free_var();
    int size = FV.size();
    FV.setminus(bound);
    if (FV.size() != size) return false;    // so there was something to remove

    match[ll] = with;
    return true;
  }

  /* constants */
  if (left->query_constant()) {
    return left == with;
  }

  /* abstractions */
  if (left->query_abstraction()) {
    if (!with->query_abstraction()) return false;

    PVariable var1 = left->query_abstraction_variable();
    PVariable var2 = with->query_abstraction_variable();
    PTerm subterm1 = left->subterm("1");
    PTerm subterm2 = with->subterm("1");

    if (!var1->query_type()->equals(var2->query_type())) return false;

    // both are abstractions with variables of the same type
    // we need to know: are their subterms equal?
    // to avoid problems, change the abstracted variable into something new
    // for both

    // 1: make a new variable (with a unique name)
    Alphabet temp_sigma;
    Environment temp_gamma = with->free_var();
    temp_gamma.combine(left->free_var());
    temp_gamma.combine(bound);
    string newname = temp_gamma.add(var1->query_type()->copy(), temp_sigma);
    bound.add(temp_gamma[newname]);

    // 2: substitute it for the abstraction variable in both subterms
    Substitution sub1, sub2;
    sub1[var1] = bound[newname];
    sub2[var2] = bound[newname];
    subterm1 = subterm1->apply_substitution(sub1);
    subterm2 = subterm2->apply_substitution(sub2);

    // 3: try to find the match :)
    bool ret = match_left(subterm1, subterm2, match, bound);

    // 4: restore the subterms to their old values!
    Substitution sub3, sub4;
    sub3[bound[newname]] = var1;
    sub4[bound[newname]] = var2;
    subterm1 = subterm1->apply_substitution(sub3);
    subterm2 = subterm2->apply_substitution(sub4);

    delete bound[newname];
    bound.remove(newname);

    return ret;
  }

  /* applications */
  if (left->query_application()) {
    if (!with->query_application()) return false;

    return match_left(left->subterm("1"), with->subterm("1"), match, bound) &&
           match_left(left->subterm("2"), with->subterm("2"), match, bound);
  }
}

bool MatchRule :: matches(PTerm s, Substitution &match) {
  Environment bound;
  return match_left(l, s, match, bound);
}

PTerm MatchRule :: apply_top(PTerm term) {
  Substitution match;
  if (!query_valid() || !matches(term, match)) return term;

  PTerm ret = r->copy();
  ret = ret->apply_substitution(match);
  term->free_memory();
  return ret;
}

bool MatchRule :: applicable_top(PTerm term) {
  Substitution match;
  return query_valid() && matches(term, match);
}

void MatchRule :: free_memory() {
  if (l != NULL) l->free_memory();
  if (r != NULL) r->free_memory();
  delete this;
}

string MatchRule :: to_string() {
  return l->to_string() + " => " + r->to_string();
}

