#include "program.h"

/* "unknown type", used for dynamic typing in the TERM_helper function
 * can be recognised because it's neither a base nor a composed type
 */
class Unknown : public Type {
  string to_string(bool brackets) {return "[unknown]";}
  PType copy() {return new Unknown;}
};

// removes all outer spaces
void TermRewriteProgram :: remove_outer_spaces(string &description) {
  while (description.length() > 0 && description[0] == ' ') {
    description = description.substr(1);
  }   
  while (description.length() > 0 && description[description.length()-1] == ' ') {
    description = description.substr(0, description.length()-1);
  }   
}

// removes all outer brackets
void TermRewriteProgram :: remove_outer_brackets(string &description) {
  // check for, and remove, outer brackets
  int n = description.length();
  if (n == 0) return;
  if (description[0] == '(' && description[n-1] == ')') {
    bool outer = true;
    int Nbrac = 1;
    for (int i = 1; i < n-1; i++) {
      if (description[i] == '(') Nbrac++;
      if (description[i] == ')') Nbrac--;
      if (Nbrac == 0) {outer = false; break;}
    }   
    if (outer) {
      description = description.substr(1, n-2);
      remove_outer_brackets(description);
    }   
  }   
}

// removes outer brackets and spaces until none are left
void TermRewriteProgram :: normalise(string &description) {
  int n = description.length()+1;

  while (description.length() != n) {
    n = description.length();
    remove_outer_spaces(description);
    remove_outer_brackets(description);
  }
}

// returns true and gives the warning if description is empty
bool TermRewriteProgram :: check_empty(string description, string warning) {
  if (description.length() == 0) {
    if (warning.length() > 0) cout << warning << endl;
    return true;
  }
  return false;
}

// return true and gives a warning if description contains illegal characters
bool TermRewriteProgram :: check_illegal(string description, bool type) {
  int n = description.length();
  for (int i = 0; i < n; i++) {
    if (description[i] == '-') {
      if (i == n-1 || description[i+1] != '>') {
        cout << "Illegal character: -" << endl;
        return true;
      }
    }
    if (description[i] == '>') {
      if (i == 0 || description[i-1] != '-') {
        cout << "Illegal character: >" << endl;
        return true;
      }
    }
    if (description[i] == '<') {
      cout << "Illegal character: <" << endl;
      return true;
    }
    if (type && (description[i] == '.' || description[i] == ',')) {
      cout << "Illegal character in type: " << description[i] << endl;
      return true;
    }
  }
  return false;
}

// returns false and gives a warning if brackets don't match
bool TermRewriteProgram :: check_correct_bracket_matching(string description) {
  int n = description.length();
  int Nbrac = 0;

  for (int i = 0; i < n; i++) {
    if (description[i] == '(') Nbrac++;
    if (description[i] == ')') {
      if (Nbrac == 0) {
        cout << "Incorrect bracket matching!" << endl;
        return false;
      }
      Nbrac--;
    }
  }
  if (Nbrac != 0) {
    cout << "Missing closing bracket!" << endl;
    return false;
  }
  return true;
}

// returns whether the string contains the given character
bool TermRewriteProgram :: contains(string description, char c) {
  int n = description.length();
  for (int i = 0; i < n; i++) {
    if (description[i] == c) return true;
  }
  return false;
}

// returns the position of the given substring, -1 if it can not be found
// everything inside brackets is ignored!
int TermRewriteProgram :: find_substring(string description, string sub) {
  int n = description.length();
  int Nbrac = 0;

  if (sub.length() == 0) return 0;

  for (int i = 0; i < n; i++) {
    if (description[i] == '(') Nbrac++;
    if (description[i] == ')') Nbrac--;
    if (description[i] == sub[0] && Nbrac == 0) {
      if (description.substr(i,sub.length()) == sub) return i;
    }
  }
  return -1;
}

// returns whether the given type has any "unknown" subtypes
bool TermRewriteProgram :: complete_type(PType type) {
  if (type->query_base()) return true;
  if (type->query_composed()) {
    return complete_type(type->query_input()) &&
           complete_type(type->query_output());
  }
  // neither base nor composed!
  return false;
}

// returns whether incomplete can be turned into complete by changing
// Unknowns into BaseTypes or ComposedTypes.
bool TermRewriteProgram :: matching_type(PType incomplete, PType complete) {
  if (incomplete->query_base()) {
    return complete->to_string() == incomplete->to_string();
  }
  if (incomplete->query_composed()) {
    if (!complete->query_composed()) return false;

    return matching_type(incomplete->query_input(), complete->query_input()) &&
           matching_type(incomplete->query_output(), complete->query_output());
  }
  // incomplete is an Unknown, which matches everything!
  return true;
}

// this function assumes description contains no illegal characters
// and has correct bracket matching
PType TermRewriteProgram :: TYPE_helper(string description) {
  normalise(description);
  if (check_empty(description, "Empty type")) return NULL;

  int arr = find_substring(description, "->");
  if (arr >= 0) {
    PType left = TYPE_helper(description.substr(0,arr));
    if (left == NULL) return NULL;
    PType right = TYPE_helper(description.substr(arr+2));
    if (right == NULL) {
      left->free_memory();
      return NULL;
    }
    return new ComposedType(left, right);
  }

  // should be a base type, or wrong if it contains spaces or brackets
  if (contains(description, ' ') ||
      contains(description, '(') ||
      contains(description, ')')) {
    cout << "Illegal: non-function type of multiple components [" << description << "]" << endl;
    return NULL;
  }

  return new BaseType(description);
}

PTerm TermRewriteProgram :: TERM_helper(string description, Alphabet
    sigma, Environment &gamma, PType expected_type, int &errcode) {

  errcode = 0;
  normalise(description);
  if (check_empty(description, "Empty term")) return NULL;

  /* If it doesn't contain spaces, brackets or dots, it's either a variable or
   * a constant.  Those are easiest!
   */
  if (!contains(description, ' ') && !contains(description, '.') &&
      !contains(description, '(') && !contains(description, ')')) {
    
    // a constant
    if (sigma.contains(description)) {
      if (!matching_type(expected_type, sigma[description]->query_type())) {
        errcode = 2;
        return NULL;
      }
      else return sigma[description];
    }

    // a known variable
    if (gamma.contains(description)) {
      if (!matching_type(expected_type, gamma[description]->query_type())) {
        errcode = 2;
        return NULL;
      }
      return gamma[description];
    }

    // if not one of the former, it's an unknown variable
    if (complete_type(expected_type)) {
      gamma[description] = new Variable(description, expected_type->copy());
      return gamma[description];
    }
    else {
      errcode = 1;
      return NULL;
    }
  }

  /* If it starts with /, it should be an abstraction.
   * We'll check whether it has the right format, and save variable, type and
   * term while doing so.
   */
  if (description[0] == '/') {
    // can this even be typed?
    if (expected_type->query_base()) {
      errcode = 2;
      return NULL;
    }    

    // needs enough characters for at least /\x:o.x
    if (description.length() < 7) {
      cout << "Illegal abstraction: " << description << "; term too short" << endl;
      return NULL;
    }

    // needs to start with lambda
    if (description[1] != '\\') {
      cout << "Illegal abstraction: " << description << "; should start with /\\" << endl;
      return NULL;
    }

    // then the variable and colon
    int colon = find_substring(description, ":");
    if (colon == -1) {
      cout << "Illegal abstraction, missing type: " << description << endl;
      return NULL;
    }
    string var = description.substr(2,colon-2);
    remove_outer_spaces(var);
    if (check_empty(var, "")) {
      cout << "Illegal abstraction: " << description << "; no variable given " << endl;
      return NULL;
    }
    if (contains(var, '.')) {
      cout << "Illegal abstraction, abstraction variable is untyped: " << description << endl;
      return NULL;
    }
    if (contains(var, '(') ||
        contains(var, ')') ||
        contains(var, '.') ||
        contains(var, ',') ||
        contains(var, ' ')) {
      cout << "Illegal abstraction: " << description << "; variable not given or illegal" << endl;
      return NULL;
    }

    // somewhere after the colon we need a dot or a comma
    int dot = find_substring(description, ".");
    int comma = find_substring(description, ",");
    if (comma != -1 && comma < dot) dot = comma;
    if (dot == -1 || dot < colon) {
      cout << "Illegal abstraction: " << description << endl;
      return NULL;
    }

    // save type and the subterm of the abstraction
    string tp = description.substr(colon+1,dot-colon-1);
    PType type = TYPE(tp);
    if (type == NULL) return NULL;
    string subterm = description.substr(dot+1);
    if (description[dot] == ',') subterm = "/\\" + subterm;

    // does the type even match?
    PType expected_type2 = expected_type;
    if (expected_type->query_composed()) {
      if (!matching_type(expected_type->query_input(), type)) {
        errcode = 2;
        type->free_memory();
        return NULL;
      }
      expected_type2 = expected_type->query_output();
    }

    // Do we already know this variable name? If so, rename.
    // Either way, add it to the environment.
    if (gamma.contains(var)) {
      string var2 = gamma.add(type, sigma, true);
      while (true) {
        int loc = subterm.find(var);
        if (loc == string::npos) break;
        else subterm.replace(loc, var.length(), var2);
      }
      var = var2;
    }
    else gamma.add(new Variable(var, type));

    // now do the work for subterm :)
    PTerm sub = TERM_helper(subterm, sigma, gamma, expected_type2, errcode);

    Variable* v = gamma[var];
    gamma.remove(var);

    if (sub == NULL) {
      delete v;
      return NULL;
    }
    else return new Abstraction(v, sub);
  }

  /* Finally, we have an application; we can either have something ending with
   * (...), or a space in it. Anything else is an illegal format.
   */
  string left, right;
  if (description[description.length()-1] == ')') {
    int Nbrac = 1;
    int i;
    for (i = description.length()-2; Nbrac != 0; i--) {
      if (description[i] == '(') Nbrac--;
      if (description[i] == ')') Nbrac++;
      if (Nbrac == 0) break;
    }
    left = description.substr(0,i);
    right = description.substr(i);
  }
  else {
    int lastspace = -1;
    for (int i = 0; i < description.length(); i++) {
      if (description[i] == ' ') lastspace = i;
      if (description[i] == ')') lastspace = i;
    }
    if (lastspace == -1) {
      cout << "Error! Illegal format: " << description << endl;
      return NULL;
    }
    left = description.substr(0,lastspace+1);
    right = description.substr(lastspace+1);
  }

  PType ex = new Unknown;
  PTerm r = TERM_helper(right, sigma, gamma, ex, errcode);
  delete ex;

  if (errcode == 2) {
    return NULL;
  }
  
  if (errcode == 1) {
    PType temptype = new ComposedType(new Unknown, expected_type->copy());
    PTerm l = TERM_helper(left, sigma, gamma, temptype, errcode);
    temptype->free_memory();

    if (errcode > 0 || l == NULL) return NULL;

    r = TERM_helper(right, sigma, gamma, l->query_type()->query_input(),
                    errcode);
    if (errcode > 0 || r == NULL) {
      l->free_memory();
      return NULL;
    }
    return new Application(l, r);

  }

  if (errcode == 0) {
    if (r == NULL) return NULL;

    PType temp = new ComposedType(r->query_type()->copy(), expected_type->copy());
    PTerm l = TERM_helper(left, sigma, gamma, temp, errcode);
    temp->free_memory();
    if (errcode > 0 || l == NULL) {
      r->free_memory();
      return NULL;
    }
    return new Application(l,r);
  }
}

PType TermRewriteProgram :: TYPE(string description) {
  if (check_illegal(description, true)) return NULL;
  if (!check_correct_bracket_matching(description)) return NULL;
  return TYPE_helper(description);
}

PTerm TermRewriteProgram :: TERM(string description, Alphabet sigma,
    Environment &gamma, PType expected_type) {

  int errcode;

  if (check_illegal(description, false)) return NULL;
  if (!check_correct_bracket_matching(description)) return NULL;

  if (expected_type == NULL) expected_type = new Unknown;

  PTerm p = TERM_helper(description, sigma, gamma, expected_type, errcode);

  if (!expected_type->query_base() && !expected_type->query_composed())
    expected_type->free_memory();

  switch(errcode) {
    case 0:
      return p;
    case 1:
      cout << "Typing failure: could not determine type of variables." << endl;
      return NULL;
    case 2:
      cout << "Typing error: term can not be typed." << endl;
      return NULL;
  }
}

