#include "term.h"
#include "alphabet.h"
#include "environment.h"
#include "substitution.h"

Term :: Term() : type(NULL) {}
Term :: Term(PType _type) : type(_type) {}

Term :: ~Term() {
  if (type != NULL) type->free_memory();
}

/* As this is virtual and the child classes overwrite it, this should
 * never be called if the classes are used correctly!
 */
string Term :: to_string(bool brackets) {
  return "[Term Error]";
}

/* Some of the children of terms are not essentially copied, like the
 * constants: they are all just pointers to the same constant in the
 * alphabet. If a child term is not of a unique kind, it should
 * overwrite this function to really create a new term.
 */
PTerm Term :: copy() {
  return this;
}

bool Term :: query_constant() {
  return false;
}

bool Term :: query_variable() {
  return false;
}

bool Term :: query_abstraction() {
  return false;
}

bool Term :: query_application() {
  return false;
}

bool Term :: query_special(string description) {
  return false;
}

bool Term :: equals(PTerm other) {
  return to_string() == other->to_string();
}

void Term :: free_memory() {
}

/* Should never be called in the inherit!
 * Returns NULL so the error will be noticed as soon as possible.
 */
PTerm Term :: apply_substitution(Substitution s) {
  return NULL;
}

PTerm Term :: subterm(string position) {
  if (position.length() == 0) return this;
  else return NULL;
}

/* Should never be called in the inherit! */
PTerm Term :: replace_subterm(PTerm subterm, string position) {
  return NULL;
}

PVariable Term :: query_abstraction_variable() {
  return NULL;
}

Environment Term :: free_var() {
  Environment a;
  return a;
}

PTerm Term :: query_head() {
  return this;
}

PType Term :: query_type() {
  return type;
}

list<string> Term :: query_positions(string startwith) {
  list<string> ret;
  ret.push_front(startwith);
  return list<string>(1, startwith);
}

string Term :: to_string_complete() {
  if (type == NULL) return to_string() + " : [Error: type unset]";
  else return to_string() + " : " + type->to_string();
}

Constant :: Constant(string _name, PType _type) : name(_name), Term(_type) {}
Constant :: Constant(const Constant& c) :name(c.name), Term(c.type->copy()) {}

string Constant :: query_name() {
  return name;
}

bool Constant :: query_constant() {
  return true;
}

bool Constant :: equals(PTerm other) {
  return dynamic_cast<PTerm>(this) == other;
}

string Constant :: to_string(bool brackets) {
  return name;
}

PTerm Constant :: apply_substitution(Substitution s) {
  return this;
}

unsigned int Constant :: arity() {
  unsigned int ret = 0;
  for (PType tmp = type; tmp->query_composed() ; tmp = tmp->query_output())
    ret++;
  return ret;
}

Variable :: Variable(string _name, PType _type) : name(_name), Term(_type) {}
Variable :: Variable(const Variable& v) :name(v.name), Term(v.type->copy()) {}

void Variable :: set_name(string _name) {
  name = _name;
}

string Variable :: query_name() {
  return name;
}

bool Variable :: query_variable() {
  return true;
}

bool Variable :: equals(PTerm other) {
  return dynamic_cast<PTerm>(this) == other;
}

string Variable :: to_string(bool brackets) {
  return name;
}

PTerm Variable :: apply_substitution(Substitution s) {
  if (s.contains(this)) return s[this]->copy();
  else return this;
}

Environment Variable :: free_var() {
  Environment a;
  a.add(this);
  return a;
}

Abstraction :: Abstraction(PVariable _var, PTerm _term)
    : var(_var), term(_term) {
  calculate_type();
}

Abstraction :: Abstraction(const Abstraction& a)
    : var(a.var), term(a.term->copy()) {}

bool Abstraction :: query_abstraction() {
  return true;
}

bool Abstraction :: equals(PTerm other) {
  if (!other->query_abstraction()) return false;
  
  Abstraction* ot = dynamic_cast<Abstraction*>(other);
  Substitution sub;
  sub[ot->var] = var;
  PTerm t = ot->term->apply_substitution(sub);
  bool ret = term->equals(t);

  Substitution back;
  back[var] = ot->var;
  ot->term = t->apply_substitution(back);

  return ret;
}

void Abstraction :: free_memory() {
  if (term != NULL) term->free_memory();
  if (var != NULL) delete var;
  delete this;
}

void Abstraction :: calculate_type() {
  PType input = var->query_type()->copy();
  PType output = term->query_type()->copy();
  type = new ComposedType(input, output);
}

string Abstraction :: to_string(bool brackets) {
  string subterm, middle, dot;
  
  subterm = term->to_string();
  dot = ". ";
  if (term->query_abstraction()) {
    subterm = subterm.substr(2);
    dot = ", ";
  }

  middle = "/\\" + (var->to_string()) + ":" +
    var->query_type()->to_string() + dot + subterm;
  if (brackets) return "(" + middle + ")";
  else return middle;
}

PTerm Abstraction :: copy() {
  Variable* newvar = new Variable(var->query_name(),var->query_type()->copy());
  Substitution sub;
  sub[var] = newvar;
  PTerm newterm = term->copy()->apply_substitution(sub);
  return new Abstraction(newvar, newterm);
}

PTerm Abstraction :: apply_substitution(Substitution s) {
  if (!s.contains(var)) term = term->apply_substitution(s);
  return this;
}

PTerm Abstraction :: subterm(string position) {
  if (position.length() == 0) return this;
  if (position[0] == '1') return term->subterm(position.substr(1));
  return NULL;
}

PTerm Abstraction :: replace_subterm(PTerm subterm, string position) {
  if (position.length() == 0 || position[0] != '1') return NULL;
  if (position.length() == 1) {
    PTerm ret = term;
    term = subterm;
    return ret;
  }
  return term->replace_subterm(subterm, position.substr(1));
}

PVariable Abstraction :: query_abstraction_variable() {
  return var;
}

Environment Abstraction :: free_var() {
  Environment FV = term->free_var();
  FV.remove(var->query_name());
  return FV;
}

list<string> Abstraction :: query_positions(string startwith) {
  list<string> ret = term->query_positions(startwith + "1");
  ret.push_front(startwith);
  return ret;
}

Application :: Application(PTerm _left, PTerm _right)
    : left(_left), right(_right) {
  calculate_type();
}

Application :: Application(const Application& a)
    : left(a.left->copy()), right(a.right->copy()) {}

bool Application :: query_application() {
  return true;
}

bool Application :: equals(PTerm other) {
  if (!other->query_application()) return false;

  Application* app = dynamic_cast<Application*>(other);
  return left->equals(app->left) && right->equals(app->right);
}

void Application :: free_memory() {
  if (left != NULL) left->free_memory();
  if (right != NULL) right->free_memory();
  delete this;
}

void Application :: calculate_type() {
  type = left->query_type()->query_output()->copy();
}

string Application :: to_string(bool brackets) {
  string middle = right->to_string(true);
  PTerm l = left;
  while (l->query_application()) {
    Application *t = dynamic_cast<Application*>(l);
    middle = t->right->to_string(true) + " " + middle;
    l = t->left;
  }
  middle = l->to_string(true) + " " + middle;

  if (brackets) return "(" + middle + ")";
  else return middle;
}

PTerm Application :: copy() {
  return new Application(left->copy(), right->copy());
}

PTerm Application :: apply_substitution(Substitution s) {
  left = left->apply_substitution(s);
  right = right->apply_substitution(s);
  return this;
}

PTerm Application :: subterm(string position) {
  if (position.length() == 0) return this;
  if (position[0] == '1') return left->subterm(position.substr(1));
  if (position[0] == '2') return right->subterm(position.substr(1));
  return NULL;
}

PTerm Application :: replace_subterm(PTerm subterm, string position) {
  if (position.length() == 0) return NULL;
  if (position[0] != '1' && position[0] != '2') return NULL;
  if (position.length() == 1) {
    PTerm ret = (position[0] == '1') ? left : right;
    if (position[0] == '1') left = subterm;
    else right = subterm;
    return ret;
  }
  if (position[0] == '1')
    return left->replace_subterm(subterm, position.substr(1));
  else
    return right->replace_subterm(subterm, position.substr(1));
}

Environment Application :: free_var() {
  Environment l = left->free_var();
  Environment r = right->free_var();
  l.combine(r);
  return l;
}

PTerm Application :: query_head() {
  return left->query_head();
}

list<string> Application :: query_positions(string startwith) {
  list<string> l = left->query_positions(startwith + "1");
  list<string> r = right->query_positions(startwith + "2");

  while (!r.empty()) {
    l.push_back(r.front());
    r.pop_front();
  }
  l.push_front(startwith);
  return l;
}

