#ifndef TERM_H
#define TERM_H

#include "type.h"
#include <list>

/**
 * These classes implement terms.
 *
 * Terms can be either constants, variables, abstractions or
 * applications, and are only referenced by pointers.  Constants and
 * variables are supposed to be unique: different occurances of the
 * same constant / variable in a term are the same pointer.  These
 * pointers are kept track of in an alphabet (for constants) or an
 * environment (for free variables).  In a bound variable the
 * abstraction that binds it keeps its own track of the variable.
 *
 * It is usually not a good idea to reuse the same term pointer in a
 * second place (other than constants and variables); instead, a
 * copy() function is provided which returns a copy of the same term;
 * only the pointers to variables and constants are unchanged.
 *
 * Do not delete a PTerm when you're done with is, just call
 * free_memory() on it.  This will not delete free variables and
 * constants!  These will have to be freed from within the alphabet /
 * environment where they are defined.  Using delete will not free
 * any subtypes at all, free_memory() will.
 *
 * Each subterm in a term carries its own type; in an abstraction
 * /\x.c, the pointer to the type of c will not be used in the type
 * of the whole abstraction.
 */

class Term;
class Variable;
class Constant;

typedef Term* PTerm;
typedef Variable* PVariable;
typedef Constant* PConstant;

class Substitution;
class Environment;

class Term {
  protected:
    PType type;

  public:
    Term();
    Term(PType _type);
    ~Term();

    virtual string to_string(bool brackets = false);
    virtual PTerm copy();
    virtual bool query_constant();
    virtual bool query_variable();
    virtual bool query_abstraction();
    virtual bool query_application();
    virtual bool query_special(string description);
      /* query_special always returns false for the terms defined
       * here; it is used to convenience different subclasses
       */
    virtual bool equals(PTerm other);
      /* returns whether this term is "equal" to the given other; that
       * is, alpha-equal under assumption of the same environment and
       * alphabet
       */
    virtual void free_memory();
      /* this does a deep free of the term, unlike delete which just
       * deletes the top symbol
       *
       * above that, it does not delete (sub)terms that are unique
       * enough to be used more than once in a term / several terms
       * (variables and constants)
       */
    virtual PTerm apply_substitution(Substitution s);
      /* this will modify the current term, and returns a pointer to
       * the result; the old pointer to this term should be forgotten,
       * not deleted
       * (it will only point to different memory if this term is a
       * substituted variable)
       */
    virtual PTerm subterm(string position);
      /* should be a string of 1s and 2s;
       * this returns an actual subterm of this term, so if anything
       * dangerous is going to happen with it, the caller should make
       * a copy() of it!
       * if there is no subterm on the given position, NULL is
       * returned
       */
    virtual PTerm replace_subterm(PTerm subterm, string position);
      /* replaces the subterm on the given position by subterm, and
       * returns a pointer to the old subterm (which is not deleted
       * or otherwise modified); returns NULL if no such position
       * exists, or the position is empty
       */
    virtual PVariable query_abstraction_variable();
      /* returns NULL if this term is not an abstraction, the
       * variable used in the abstraction otherwise
       */
    virtual Environment free_var();
      /* returns an environment containing all the free variables in
       * the term
       */
    virtual PTerm query_head();
      /* returns the head term (where the head term of an application
       * s0 s1 ... sn is s0 as long as s0 is not an application); if
       * the term is beta-normalised, this is either a variable or a
       * constant, otherwise it might also be an abstraction
       *
       * note: this returns a direct subterm, not a copy; do not just
       * delete (or free_memory)
       */
    virtual list<string> query_positions(string startwith = "");
      /* returns a list containing all positions in this term
       * startwith is placed in front of all strings in the list (and
       * is used for the recursion)
       */

    PType query_type();
      /* returns the type of this term, do not delete or free */
    string to_string_complete();
      /* returns the textual description with type */
};

class Constant : public Term {
  private:
    string name;

  public:
    Constant(string _name, PType _type);
    Constant(const Constant& c);

    string query_name();
    bool query_constant();
    bool equals(PTerm other);
    string to_string(bool brackets = false);
    PTerm apply_substitution(Substitution s);

    unsigned int arity();
};

class Variable : public Term {
  private:
    string name;

  public:
    Variable(string _name, PType _type);
    Variable(const Variable& v);

    void set_name(string _name);
    string query_name();
    bool query_variable();
    bool equals(PTerm other);
    string to_string(bool brackets = false);
    PTerm apply_substitution(Substitution s);
    Environment free_var();
};

class Abstraction : public Term {
  private:
    PVariable var;
    PTerm term;

    void calculate_type();

  public:
    Abstraction(PVariable _var, PTerm _term);
    Abstraction(const Abstraction& a);

    bool query_abstraction();
    bool equals(PTerm other);
    void free_memory();
    string to_string(bool brackets = false);
    PTerm copy();
    PTerm apply_substitution(Substitution s);
    PTerm subterm(string position);
    PTerm replace_subterm(PTerm subterm, string position);
    PVariable query_abstraction_variable();
    Environment free_var();
    list<string> query_positions(string startwith = "");
};

class Application : public Term {
  private:
    PTerm left, right;

    void calculate_type();

  public:
    Application(PTerm _left, PTerm _right);
    Application(const Application& a);

    bool query_application();
    bool equals(PTerm other);
    void free_memory();
    string to_string(bool brackets = false);
    PTerm copy();
    PTerm apply_substitution(Substitution s);
    PTerm subterm(string position);
    PTerm replace_subterm(PTerm subterm, string position);
    Environment free_var();
    PTerm query_head();
    list<string> query_positions(string startwith = "");
};

#endif

