#include "type.h"

/* As this is virtual and all the child classes are required to
 * overwrite it, this should never be called if the classes are used
 * correctly.
 * It retuns NULL rather than this so the error will occur at the
 * point where it is used, rather than at a completely different
 * place in the program where the original type is freed and the copy
 * used.
 */
PType Type :: copy() {
  return NULL;
}

/* Should also never be called. */
string Type :: to_string(bool brackets) {
  return "[Type Error]";
}

void Type :: free_memory() {
  delete this;
}

bool Type :: query_base() {
  return false;
}

bool Type :: query_composed() {
  return false;
}

PType Type :: query_input() {
  return NULL;
}

PType Type :: query_output() {
  return this;
}

bool Type :: equals(PType other) {
  return to_string() == other->to_string();
}

PType Type :: collapse() {
  return new BaseType("o");
}

PType BaseType :: copy() {
  return new BaseType(name);
}

string BaseType :: to_string(bool brackets) {
  return name;
}

bool BaseType :: query_base() {
  return true;
}

PType ComposedType :: copy() {
  if (left == NULL || right == NULL) return NULL;

  PType l = left->copy();
  PType r = right->copy();
  return new ComposedType(l, r);
}

string ComposedType :: to_string(bool brackets) {
  if (left == NULL || right == NULL) return "[Broken Type]";

  string middle = left->to_string(true) + " -> " + right->to_string(false);
  if (brackets) return "(" + middle + ")";
  else return middle;
}

void ComposedType :: free_memory() {
  if (left != NULL) left->free_memory();
  if (right != NULL) right->free_memory();
  delete this;
}

bool ComposedType :: equals(PType other) {
  if (!other->query_composed()) return false;

  if (left == NULL || right == NULL) return false;

  return left->equals(other->query_input()) &&
         right->equals(other->query_output());
}

bool ComposedType :: query_composed() {
  return true;
}

PType ComposedType :: query_input() {
  return left;
}

PType ComposedType :: query_output() {
  return right;
}

PType ComposedType :: collapse() {
  return new ComposedType(left->collapse(), right->collapse());
}
