#ifndef TYPE_H
#define TYPE_H

#include <cstdlib>
#include <string>
#include <iostream>
using namespace std;

/**
 * These classes implement typing.
 *
 * Under normal usage, Types are only referenced by pointers.  The
 * user (where "user" would be something like a term or the rest of
 * the program) keeps a PType.
 *
 * As PTypes are pointers, it is usually not a good idea to reuse the
 * same type for something else.  Instead, a copy() function is
 * provided which creates a copy of the type.
 *
 * Do not delete a PType when you are done with it, just call
 * free_memory() on it.  Delete will not free up subtypes, using
 * free_memory() will.
 */

class Type;
typedef Type* PType;

class Type {
  public:
    virtual PType copy();
    virtual string to_string(bool brackets = false);
    virtual void free_memory();

    virtual bool equals(PType other);
    virtual bool query_base();
    virtual bool query_composed();
      // returns whether this and other are the same type
    virtual PType query_input();
      // this will return NULL if this is not a composed type!
      // otherwise, it will return a subtype, NOT a copy of a subtype
    virtual PType query_output();
      // this will just return this object if it is not a composed
      // type
      // it will always return a subtype, NOT a copy of a subtype
    virtual PType collapse();
      // returns this type but with all base types (and other non
      // composed types) replaced by the base type o;
      // this creates a new PType, it doesn't affect the current one
};

class BaseType : public Type {
  private:
    string name;

  public:
    BaseType(string _name) : name(_name) {}

    PType copy();
    string to_string(bool brackets = false);
    bool query_base();
};

class ComposedType : public Type {
  private:
    PType left, right;

  public:
    ComposedType(PType _left, PType _right) : left(_left), right(_right) {}

    PType copy();
    string to_string(bool brackets = false);
    void free_memory();
    bool equals(PType other);
    bool query_composed();
    PType query_input();
    PType query_output();
    PType collapse();
};

/* use these to easily create types */

#endif

